package logging

import (
	"fmt"
	"github.com/pkg/errors"
	log "github.com/sirupsen/logrus"
	"runtime"
	"runtime/debug"
	"strings"
	"sync"
	"time"
)

var (
	instance *log.Entry = nil
	once     sync.Once
)

const logLayout = "[func %s(), file:%s:%d] : %s"

type Level uint32

const (
	// PanicLevel level, highest level of severity. Logs and then calls panic with the
	// message passed to Debug, Info, ...
	PanicLevel Level = iota
	// FatalLevel level. Logs and then calls `os.Exit(1)`. It will exit even if the
	// logging level is set to Panic.
	FatalLevel
	// ErrorLevel level. Logs. Used for errors that should definitely be noted.
	// Commonly used for hooks to send errors to an error tracking service.
	ErrorLevel
	// WarnLevel level. Non-critical entries that deserve eyes.
	WarnLevel
	// InfoLevel level. General operational entries about what's going on inside the
	// application.
	InfoLevel
	// DebugLevel level. Usually only enabled when debugging. Very verbose logging.
	DebugLevel
)

func Get() *LoggerHandler {
	once.Do(func() {
		instance = log.WithFields(log.Fields{})
		formatter := new(myFormatter)
		formatter.textFormatter.FullTimestamp = false
		formatter.textFormatter.DisableTimestamp = true
		formatter.textFormatter.ForceColors = true
		formatter.textFormatter.DisableColors = false
		log.SetFormatter(formatter)
	})

	return &LoggerHandler{
		logger: instance,
	}
}

func GetString(arr []byte) string {
	if arr == nil || len(arr) == 0 {
		return "NULL"
	} else {
		return string(arr)
	}
}

type LoggerHandler struct {
	logger *log.Entry
}

func (obj *LoggerHandler) SetLevel(level Level) {
	log.SetLevel(log.Level(level))
}

func (obj *LoggerHandler) Infof(format string, a ...interface{}) {
	pc, fn, line, _ := runtime.Caller(1)

	fullMsg := fmt.Sprintf(logLayout,
		colored(cut(runtime.FuncForPC(pc).Name()), fgBlue, faint),
		colored(cut(fn), fgMagenta, faint),
		line,
		colored(format, fgBlack, faint))

	obj.logger.Printf(fullMsg, a...)
}

func (obj *LoggerHandler) Debugf(format string, a ...interface{}) {
	pc, fn, line, _ := runtime.Caller(1)

	fullMsg := fmt.Sprintf(logLayout,
		colored(cut(runtime.FuncForPC(pc).Name()), fgBlue, faint),
		colored(cut(fn), fgMagenta, faint),
		line,
		colored(format, fgBlack, faint))

	obj.logger.Debugf(fullMsg, a...)
}

func (obj *LoggerHandler) Warnf(format string, a ...interface{}) {
	pc, fn, line, _ := runtime.Caller(1)

	fullMsg := fmt.Sprintf(logLayout,
		colored(cut(runtime.FuncForPC(pc).Name()), fgBlue, faint),
		colored(cut(fn), fgMagenta, faint),
		line,
		colored(format, fgBlack, faint))

	obj.logger.Warnf(fullMsg, a...)
}

func (obj *LoggerHandler) Errorf(format string, a ...interface{}) error {
	pc, fn, line, _ := runtime.Caller(1)

	fullMsg := fmt.Sprintf(logLayout+"\n%s",
		colored(cut(runtime.FuncForPC(pc).Name()), fgBlue, faint),
		colored(cut(fn), fgMagenta, faint),
		line,
		colored(fmt.Sprintf(format, a...), fgBlack, faint),
		colored(string(debug.Stack()), fgRed, faint))

	obj.logger.Error(fullMsg)

	return errors.New(fmt.Sprintf(format, a))
}

func (obj *LoggerHandler) Info(msg string) {
	pc, fn, line, _ := runtime.Caller(1)

	fullMsg := fmt.Sprintf(logLayout,
		colored(cut(runtime.FuncForPC(pc).Name()), fgBlue, faint),
		colored(cut(fn), fgMagenta, faint),
		line,
		colored(msg, fgBlack, faint))

	obj.logger.Printf(fullMsg)
}

func (obj *LoggerHandler) Debug(msg string) {
	pc, fn, line, _ := runtime.Caller(1)

	fullMsg := fmt.Sprintf(logLayout,
		colored(cut(runtime.FuncForPC(pc).Name()), fgBlue, faint),
		colored(cut(fn), fgMagenta, faint),
		line,
		colored(msg, fgBlack, faint))

	obj.logger.Debugf(fullMsg)
}

func (obj *LoggerHandler) Warning(msg string) {
	pc, fn, line, _ := runtime.Caller(1)

	fullMsg := fmt.Sprintf(logLayout,
		colored(cut(runtime.FuncForPC(pc).Name()), fgBlue, faint),
		colored(cut(fn), fgMagenta, faint),
		line,
		colored(msg, fgBlack, faint))

	obj.logger.Warnf(fullMsg)
}

func (obj *LoggerHandler) Error(msg string) error {
	pc, fn, line, _ := runtime.Caller(1)

	fullMsg := fmt.Sprintf(logLayout,
		colored(cut(runtime.FuncForPC(pc).Name()), fgBlue, faint),
		colored(cut(fn), fgMagenta, faint),
		line,
		colored(msg, fgBlack, faint))

	obj.logger.Errorf(fullMsg+"\n%s", string(debug.Stack()))
	return errors.New(fmt.Sprintf(msg))
}

type myFormatter struct {
	textFormatter log.TextFormatter
}

func (obj *myFormatter) Format(entry *log.Entry) ([]byte, error) {
	if b, err := obj.textFormatter.Format(entry); err != nil {
		return nil, err
	} else {
		s := string(b)
		s = strings.Replace(s, "\\n", "\n", len(s))
		s = strings.Replace(s, "\\t", "\t", len(s))
		s = strings.Replace(s, "\\033", "\033", len(s))
		s = strings.Replace(s, "ERRO", "ERROR", len(s))
		s = strings.Replace(s, "DEBU", "DEBUG", len(s))
		s = strings.Replace(s, "INFO", "INFO ", len(s))
		s = formatDate(entry.Time) + "  " + s
		return []byte(s), nil
	}
}

func cut(path string) string {
	idx := strings.LastIndex(path, "/")
	if idx == -1 || idx+1 == len(path) {
		return path
	} else {
		return path[idx+1:]
	}
}

func formatDate(t time.Time) string {
	layout := "2006-01-02 03:04:05.000"
	return t.Format(layout)
}

func colored(word string, color attributeColor, style attributeStyle) string {
	return fmt.Sprintf("%s[%d;%dm%s%s[0m", escape, color, style, word, escape)
}
