package market

import (
	"gitlab.smartblocklab.com/smartcontracts/common/logging"
	"strings"
	"gitlab.smartblocklab.com/smartcontracts/common/nums"
)

func (obj *PlainMarketOrder) MarketOrder(sessionID int64) (*MarketOrder, error) {
	if err := obj.Validate(); err != nil {
		return nil, err
	}
	buyAsset := nums.AssetFrom(obj.BuyAmount.CurrencyType)
	sellAsset := nums.AssetFrom(obj.SellPrice.CurrencyType)
	currencyPair := NewCurrencyPairFromAssets(buyAsset, sellAsset)
	sellPrice := obj.SellPrice.Number()
	buyAmount := obj.BuyAmount.Number()
	filledAmount := obj.FilledAmount.Number()
	orderStatus := obj.OrderStatus

	options := Options{
		StopPrice:  obj.StopPrice.Number(),
		LimitPrice: obj.LimitPrice.Number(),
	}

	return CreateOrder(obj.UserID, obj.AccountID, sessionID, obj.OperationID, orderStatus, currencyPair, sellPrice, buyAmount, filledAmount, obj.Timestamp, options), nil
}

func (obj *PlainMarketOrder) Validate() error {
	if len(obj.OperationID) == 0 {
		return logging.Get().Error("PlainMarketOrder must to have operationID")
	}

	if len(obj.UserID) == 0 {
		return logging.Get().Error("PlainMarketOrder must to have userID")
	}

	if obj.StopPrice.CurrencyType != nums.ZeroAsset && obj.StopPrice.CurrencyType != obj.SellPrice.CurrencyType {
		return logging.Get().Errorf("StopPrice and SellPrice assets are different: %s vs. %s", obj.StopPrice.CurrencyType, obj.SellPrice.CurrencyType)
	}

	if obj.LimitPrice.CurrencyType != nums.ZeroAsset && obj.LimitPrice.CurrencyType != obj.SellPrice.CurrencyType {
		return logging.Get().Errorf("LimitPrice and SellPrice assets are different: %s vs. %s", obj.StopPrice.CurrencyType, obj.SellPrice.CurrencyType)
	}

	return nil
}

func (obj *UIMarketOrder) Validate() error {
	if len(obj.AccountID) == 0 {
		return logging.Get().Error("Order account id is empty")
	}

	if len(obj.LimitPrice.Amount) == 0 {
		return logging.Get().Error("Order LimitPrice.Amount is empty")
	}

	if len(obj.StopPrice.Amount) == 0 {
		return logging.Get().Error("Order StopPrice.Amount is empty")
	}

	if len(obj.Price.Amount) == 0 {
		return logging.Get().Error("Order Price.Amount is empty")
	}

	if len(obj.LimitPrice.CurrencyType) == 0 {
		return logging.Get().Error("Order LimitPrice.CurrencyType is empty")
	}

	if len(obj.StopPrice.CurrencyType) == 0 {
		return logging.Get().Error("Order StopPrice.CurrencyType is empty")
	}

	if len(obj.Price.CurrencyType) == 0 {
		return logging.Get().Error("Order Price.CurrencyType is empty")
	}

	if obj.Price.CurrencyType == obj.Amount.CurrencyType {
		return logging.Get().Error("Sell currency type and Buy currency type is same")
	}

	return nil
}

func (obj *MarketOrder) CheckTriggers(currentPrice *nums.CurrencyNumber) bool {

	// сигнализируем, что хотим закрыть, если:
	if obj.OrderType == MARKET {
		// Готовы купить по любой цене
		return true
	} else if obj.OrderType == STOP|LIMIT {
		// Цена покупки стала выше максимально допустимой, и поэтому придется покупать (плохой исход)
		logging.Get().Debugf("StopLimit: OrderID=%s, currentPrice=%s > stopPrice=%s ?", obj.ObjectKey(), currentPrice.String(), obj.StopPrice.String())

		if currentPrice.GreaterEq(obj.StopPrice) {
			logging.Get().Debugf("StopLimit: OrderID=%s, Stop reached", obj.ObjectKey())
			return true
		}
		// Цена покупки стала ниже планируемой, и поэтому надо покупать (хороший исход)
		logging.Get().Debugf("StopLimit: OrderID=%s, currentPrice=%s < limitPrice=%s ?", obj.ObjectKey(), currentPrice.String(), obj.LimitPrice.String())

		if currentPrice.LessEq(obj.LimitPrice) {
			logging.Get().Debugf("StopLimit: OrderID=%s, Limit reached", obj.ObjectKey())
			return true
		}
	} else if obj.OrderType == STOP {
		// Цена покупки стала выше максимально допустимой, и поэтому придется покупать (плохой исход)
		logging.Get().Debugf("Stop: OrderID=%s, currentPrice=%s > stopPrice=%s ?", obj.ObjectKey(), currentPrice.String(), obj.StopPrice.String())
		if currentPrice.GreaterEq(obj.StopPrice) {
			logging.Get().Debugf("Stop: OrderID=%s, Stop reached", obj.ObjectKey())
			return true
		}
	} else if obj.OrderType == LIMIT {
		// Цена покупки стала ниже планируемой, и поэтому надо покупать (хороший исход)
		logging.Get().Debugf("Limit: OrderID=%s, currentPrice=%s < limitPrice=%s ?", obj.ObjectKey(), currentPrice.String(), obj.LimitPrice.String())
		if currentPrice.LessEq(obj.LimitPrice) {
			logging.Get().Debugf("Limit: OrderID=%s, Limit reached", obj.ObjectKey())
			return true
		}
	} else {
		// Цена спроса соответствует цене предложения
		logging.Get().Debugf("Base: OrderID=%s, currentPrice=%s == price=%s ?", obj.ObjectKey(), currentPrice.String(), obj.Price.String())

		if obj.Price.Equals(currentPrice) {
			logging.Get().Debugf("Base: OrderID=%s, Price reached, price=%s", obj.ObjectKey(), obj.Price.String())
			return true
		}
	}

	return false
}

func (obj *MarketOrder) OpenAmount() *nums.CurrencyNumber {
	return obj.Amount.Sub(obj.FilledAmount)
}

func (obj *MarketOrder) Status() OrderStatus {
	if obj.OpenAmount().Equals(nums.Zero()) {
		return OrderFilled
	}

	return obj.StatusInternal
}

func (obj *MarketOrder) SetStatus(status OrderStatus) {
	obj.StatusInternal = status
}

func (obj *MarketOrder) Plain() PlainMarketOrder {
	res := PlainMarketOrder{
		UserID:       obj.UserID,
		OperationID:  obj.OperationID,
		SessionID:    obj.SessionID,
		AccountID:    obj.AccountID,
		SellPrice:    obj.Price.Plain(),
		DocumentType: PlainOrder,
		OrderStatus:  obj.Status(),
		BuyAmount:    obj.Amount.Plain(),
		FilledAmount: obj.FilledAmount.Plain(),
		StopPrice:    obj.StopPrice.Plain(),
		LimitPrice:   obj.LimitPrice.Plain(),
		Timestamp:    obj.Timestamp,
	}

	return res
}

func CreateOrder(userID string, accountId string, sessionID int64, operationID string, status OrderStatus, pair *CurrencyPair, sellPrice *nums.CurrencyNumber,
	buyAmount *nums.CurrencyNumber, filledAmount *nums.CurrencyNumber, timestamp int64, options Options) *MarketOrder {
	var orderType OrderType
	if sellPrice.Equals(nums.Zero()) {
		if options.StopPrice.Equals(nums.Zero()) && options.LimitPrice.Equals(nums.Zero()) {
			orderType |= MARKET

		} else {
			if options.StopPrice != nil && options.StopPrice.Greater(nums.Zero()) {
				orderType |= STOP
				sellPrice = options.StopPrice
			}

			if options.LimitPrice != nil && options.LimitPrice.Greater(nums.Zero()) {
				orderType |= LIMIT

				sellPrice = options.LimitPrice
			}
		}
	}

	return &MarketOrder{
		SessionID:      sessionID,
		UserID:         userID,
		OperationID:    operationID,
		StatusInternal: status,
		AccountID:      accountId,
		CurrencyPair:   pair,
		OrderType:      orderType,
		Price:          sellPrice,
		Amount:         buyAmount,
		FilledAmount:   filledAmount,
		Timestamp:      timestamp,
		StopPrice:      options.StopPrice,
		LimitPrice:     options.LimitPrice,
	}
}

func NewCurrencyPairFromString(raw string) *CurrencyPair {
	res := strings.Split(raw, "/")
	return &CurrencyPair{BuyAsset: nums.AssetFrom(res[0]), SellAsset: nums.AssetFrom(res[1])}
}

func NewCurrencyPairFromAssets(buy *nums.Asset, sell *nums.Asset) *CurrencyPair {
	return &CurrencyPair{BuyAsset: buy, SellAsset: sell}
}

func (obj *CurrencyPair) Equals(pair *CurrencyPair) bool {
	return obj.SellAsset == pair.SellAsset && obj.BuyAsset == pair.BuyAsset
}

func (obj *CurrencyPair) ReversedEquals(pair *CurrencyPair) bool {
	return obj.BuyAsset == pair.SellAsset && obj.SellAsset == pair.BuyAsset
}

func (obj *CurrencyPair) HashKey() string {
	return obj.BuyAsset.String() + "/" + obj.SellAsset.String()
}

func (obj *CurrencyPair) ReversedHashKey() string {
	return obj.SellAsset.String() + "/" + obj.BuyAsset.String()
}

func (obj *CurrencyPair) String() string {
	return obj.BuyAsset.String() + "/" + obj.SellAsset.String()
}
