package checker

import (
	"crypto/sha256"
	"encoding/hex"
	"encoding/json"
	"errors"
	"fmt"
	log "github.com/sirupsen/logrus"
	"gitlab.smartblocklab.com/smartcontracts/sc-security/actions"
	"gitlab.smartblocklab.com/smartcontracts/sc-security/requests"
	"gitlab.smartblocklab.com/smartcontracts/sc-security/tools"
	"strings"
)

var (
	logger = log.WithFields(log.Fields{})
)

type InternalUser struct {
	UserID string          `json:"user_id"`
	Roles  []actions.Roles `json:"roles"`
}

func CheckAccessParameterByArgsAndFunctionName(args []string, chaincodeName string, functionName string, stub tools.ChaincodeStubInterfaceAdapter) (*requests.Parameter, *requests.Parameter, error) {
	preferences, err := FindPreferencesByFunctionName(chaincodeName, functionName)
	if err != nil {
		logger.Error(err)
		return nil, nil, err;
	}
	logger.Infof("preferences.FunctionName: %s", preferences.FunctionName)
	logger.Infof("preferences.ChaincodeName: %s", preferences.ChaincodeName)
	logger.Infof("preferences.ParameterType: %s", preferences.ParameterType)
	return CheckAccessParameter(stub, args, preferences)
}

func firstLetterToLowerCase(str string) string {
	firstLetter := strings.ToLower(str[0:1])
	replaced := firstLetter + str[1:]

	return replaced
}

func FindPreferencesByFunctionName(chaincodeName string, functionName string) (actions.SecurityPreference, error) {
	logger.Infof("chaincodeName: %s", chaincodeName)
	logger.Infof("functionName: %s", functionName)
	chaincodeNameLower := firstLetterToLowerCase(chaincodeName)
	functionNameLower := firstLetterToLowerCase(functionName)

	for _, v := range actions.SecurityPreferenceArray {
		if v.ChaincodeName == chaincodeNameLower && v.FunctionName == functionNameLower {
			logger.Infof("found SecurityPreference: %s", v)
			return v, nil
		}
	}
	return actions.SecurityPreference{}, errors.New("not found")
}

/**
@return UserParams, SystemParams, error
*/
func CheckAccessParameter(stub tools.ChaincodeStubInterfaceAdapter, args []string, preferences actions.SecurityPreference) (*requests.Parameter, *requests.Parameter, error) {
	logger.Info(fmt.Sprintf("CheckAccess: chaincode=%s, method=%s", preferences.ChaincodeName, preferences.FunctionName))

	if len(args) != 1 {
		errMsg := fmt.Sprintf("CheckAccess: Incorrect number of arguments, expecting %d, while got %d", 1, len(args))
		logger.Error(errMsg)
		return nil, nil, errors.New(errMsg)
	}

	logger.Infof("FullSignedRequest: %s", args[0])
	var signedMessage requests.FullSignedRequest
	if err := json.Unmarshal([]byte(args[0]), &signedMessage); err != nil {
		return nil, nil, err
	}

	var userParam *requests.Parameter
	var systemParam *requests.Parameter
	var err error

	if preferences.ParameterType == actions.UserParameter || preferences.ParameterType == actions.ComplexParameter {
		logger.Info(fmt.Sprintf("CheckAccess: user method: %s.%s", preferences.ChaincodeName, preferences.FunctionName))
		// verify user params

		if userParam, err = verifyParameter(stub, preferences, signedMessage.Parameters.UserParams, signedMessage.UserSignature); err != nil {
			logger.Error(fmt.Sprintf("CheckAccess: verification failed: %s", err.Error()))
			return nil, nil, fmt.Errorf("CheckAccess: verification failed: %s ", err.Error())
		}

		if !checkRole(stub, userParam, preferences) {
			logger.Error(fmt.Sprintf("CheckAccess: Role checking failed"))
			return nil, nil, errors.New("CheckAccess: role checking failed: user has no role")
		}

		if preferences.ParameterType == actions.UserParameter {
			log.Info(fmt.Sprintf("CheckAccess: [UserParameter] Access granted for user '%s'", userParam.UserID))
			return userParam, nil, nil
		}

		// verify system params
		if systemParam, err = verifyParameter(stub, preferences, signedMessage.Parameters.SystemParams, signedMessage.SystemSignature); err != nil {
			logger.Error(fmt.Sprintf("CheckAccess: verification failed: %s", err.Error()))
			return nil, nil, fmt.Errorf("CheckAccess: verification failed: %s ", err.Error())
		}

		if !checkRole(stub, systemParam, preferences) {
			return nil, nil, errors.New("CheckAccess: access denied: system user has no role")
		}

		log.Info(fmt.Sprintf("CheckAccess: chaincode=%s, method=%s. [UserParameter, ServiceParameter] Access granted for user '%s' and system user '%s'",
			preferences.ChaincodeName, preferences.FunctionName, userParam.UserID, systemParam.UserID))

		return userParam, systemParam, nil

	} else if preferences.ParameterType == actions.ServiceParameter {
		logger.Info(fmt.Sprintf("CheckAccess: service method: %s.%s", preferences.ChaincodeName, preferences.FunctionName))
		// verify system params
		if systemParam, err = verifyParameter(stub, preferences, signedMessage.Parameters.SystemParams, signedMessage.SystemSignature); err != nil {
			logger.Error(fmt.Sprintf("CheckAccess: verification failed: %s", err.Error()))
			return nil, nil, fmt.Errorf("CheckAccess: verification failed: %s ", err.Error())
		}

		if !checkRole(stub, systemParam, preferences) {
			return nil, nil, errors.New("CheckAccess: access denied: system user has no role")
		}

		logger.Info(fmt.Sprintf("CheckAccess: chaincode=%s, method=%s. [ServiceParameter] Access granted for system user '%s'", preferences.ChaincodeName, preferences.FunctionName, systemParam.UserID))

		return nil, systemParam, nil

	} else if preferences.ParameterType == actions.ServiceParameter|actions.UnsecuredParameter {
		logger.Info(fmt.Sprintf("CheckAccess: unsecured service method: %s.%s", preferences.ChaincodeName, preferences.FunctionName))
		if err = json.Unmarshal(signedMessage.Parameters.SystemParams, &systemParam); err != nil {
			logger.Error(fmt.Sprintf("CheckAccess: Unmarshaling error: %s", err.Error()))
			return nil, nil, err
		}
		logger.Info(fmt.Sprintf("CheckAccess: chaincode=%s, method=%s. [Unsecured service Parameter] Access granted for system user '%s'", preferences.ChaincodeName, preferences.FunctionName, systemParam.UserID))

		return nil, systemParam, nil

	} else if preferences.ParameterType == actions.UserParameter|actions.UnsecuredParameter {
		logger.Info(fmt.Sprintf("CheckAccess: unsecured user method: %s.%s", preferences.ChaincodeName, preferences.FunctionName))
		if err = json.Unmarshal(signedMessage.Parameters.UserParams, &userParam); err != nil {
			logger.Error(fmt.Sprintf("CheckAccess: Unmarshaling error: %s", err.Error()))
			return nil, nil, err
		}
		logger.Info(fmt.Sprintf("CheckAccess: chaincode=%s, method=%s. [Unsecured user Parameter] Access granted for system user '%s'", preferences.ChaincodeName, preferences.FunctionName, systemParam.UserID))

		return userParam, nil, nil

	} else {
		return nil, nil, fmt.Errorf("CheckAccess: Unknown ParameterType: %d ", preferences.ParameterType)
	}
}

func CheckAccess(stub tools.ChaincodeStubInterfaceAdapter, args []string, preferences actions.SecurityPreference) (json.RawMessage, json.RawMessage, error) {
	userParameter, systemParameter, err := CheckAccessParameter(stub, args, preferences)

	var systemParam json.RawMessage = nil
	if systemParameter != nil {
		systemParam = systemParameter.Param
	}

	var userParam json.RawMessage = nil
	if userParameter != nil {
		userParam = userParameter.Param
	}
	return userParam, systemParam, err
}

func verifyParameter(stub tools.ChaincodeStubInterfaceAdapter, preferences actions.SecurityPreference, rawParameter json.RawMessage, sig requests.UserSignature) (*requests.Parameter, error) {
	logger.Info(fmt.Sprintf("VerifyParameter: ChaincodeName=%s, method=%s, parameter=%s", preferences.ChaincodeName,
		preferences.FunctionName, string(rawParameter)))

	var parameter requests.Parameter
	if err := json.Unmarshal(rawParameter, &parameter); err != nil {
		logger.Error(fmt.Sprintf("VerifyParameter: Unmarshaling error: %s", err.Error()))
		return nil, err
	}

	if preferences.GetActionID() != parameter.ActionId {
		return nil, errors.New(fmt.Sprintf("VerifyParameter: Action Ids are not equal – called id: %d, caller id: %d",
			preferences.GetActionID(), parameter.ActionId))
	}

	hasher := sha256.New()
	hasher.Write(rawParameter)
	digest := hasher.Sum(nil)
	strDigest := hex.EncodeToString(digest)

	arg := requests.VerifyUserSignatureArg{
		UserID:        parameter.UserID,
		Digest:        strDigest,
		UserSignature: sig,
	}

	logger.Info(fmt.Sprintf("VerifyParameter: Invoke verifyUserSignature method"))
	resp := tools.InvokeChaincode_VerifyUserSignature(stub, arg)

	if resp.Status != 200 {
		msg := fmt.Sprintf("VerifyParameter: Invoke failed: %s ", resp.Message)
		log.Error(msg)
		return nil, fmt.Errorf(msg)
	} else {
		return &parameter, nil
	}
}

func checkRole(stub tools.ChaincodeStubInterfaceAdapter, parameter *requests.Parameter, preferences actions.SecurityPreference) bool {
	log.Infof("checkRole: tools.InvokeChaincode_GetUser %s", parameter.UserID)

	resp := tools.InvokeChaincode_GetUser(stub, parameter.UserID)
	log.Infof("checkRole: response status: %d", resp.Status)
	log.Infof("checkRole: payload: %s", string(resp.Payload))
	log.Infof("checkRole: message: %s", string(resp.Message))

	if resp.Status != 200 {
		log.Error(fmt.Sprintf("checkRole: Invoke chaincode 'GetUser' failed, with response status: %s, message: %s, payload: %s",
			resp.Status, resp.Message, resp.Payload))
		return false

	} else {
		var user InternalUser
		if err := json.Unmarshal(resp.Payload, &user); err != nil {
			log.Error(fmt.Sprintf("checkRole: Checking roles failed, can't unmarshal user: %s ", resp.Message))
			return false
		}

		logger.Info(fmt.Sprintf("checkRole: ChaincodeName=%s, method=%s, user=%s", preferences.ChaincodeName, preferences.FunctionName, string(resp.Payload)))

		if parameter.UserID == user.UserID {
			logger.Info("checkRole: Found same Parameter UserID and Fetched UserID")

			if len(user.Roles) == 0 && len(preferences.Roles) == 0 {
				return true
			}

			for _, userRole := range user.Roles {
				for _, methodRole := range preferences.Roles {
					if userRole == methodRole {
						logger.Info("checkRole: Role is matched")
						return true
					}
				}
			}
		}

		logger.Error("checkRole: User has no any suitable role")
		return false
	}
}
